       
        /*   
         Arduino code for simple dashboard with TM1638
        */

        // Serial parser for Interface - Output : M<Dash2>S<Dash1>G<Dash3>
        // Serial extended parser with kart racing pro : S<Dash1>M<Dash2>G<Dash3>W<Dash4>O<Dash5>F<Dash7>
        // Serial parser for Shutdown - Output : Z00
        // M : rpm              (16bit - hexadecimal output)
        // S : speed            (16bit - hexadecimal output)
        // G : gear             (16bit - hexadecimal output)
        // O : oil temp         (16bit - hexadecimal output)
        // W : water temp       (16bit - hexadecimal output)
        // F : fuel             (16bit - hexadecimal output)
        // L : lap              (16bit - hexadecimal output)
        // P : position         (16bit - hexadecimal output)
        // Z : clear            (to clear display and max rpm value)
        //
        // for dashboard defined in TM1638.h
        // Pin D3 : STB0 - strobe
        // Pin D4 : CLK - clock
        // Pin D5 : DIO - data

        ////////////////////////////////////////////////////////////////////////////////
        // DashBoard include
        ////////////////////////////////////////////////////////////////////////////////   
        // module can be downloaded from http://forum.arduino.cc/index.php?topic=190472.0
        // (search TM1638_v1.02.zip)
        #include <TM1638.h> 
        #include <TM1638Font.h>
        TM1638 module;
        
        // buttons TM1638 (left to right)
        #define btnSpeed      0b00000001 // 1
        #define btnGear       0b00000010 // 2
        #define btnRPM        0b00000100 // 4
        #define btnOIL        0b00001000 // 8
        #define btnWater      0b00010000 // 16
        #define btnFuel       0b00100000 // 32
        #define btnLAP        0b01000000 // 64
        #define btnPosition   0b10000000 // 128

        ////////////////////////////////////////////////////////////////////////////////
        // EEPROM include
        ////////////////////////////////////////////////////////////////////////////////          
        #include "EEPROM.h"

        ////////////////////////////////////////////////////////////////////////////////
        // variable init
        ////////////////////////////////////////////////////////////////////////////////          
        int i;
        unsigned int carspeed;      // holds the speed data (0-65535 size)
        unsigned int gear;          // holds gear value data (0-65535 size)
        unsigned int rpm;           // holds the rpm data (0-65535 size)
        unsigned int oil;           // holds the oil temp data (0-65535 size)
        unsigned int water;         // holds the water temp data (0-65535 size)
        unsigned int fuel;          // holds the fuel data (0-65535 size)
        unsigned int lap;           // holds the lap data (0-65535 size)
        unsigned int posit;         // holds the position data (0-65535 size)
        
        unsigned int rpmleds;          // holds the 8 leds values
        unsigned int rpmmax = 5000;    // retrieves from rpm live with minimum setting on initialization and Shutdown - Output
        unsigned int beforeButton = 0; // holds the previous pressed button        
        unsigned int valButton = 0;    // retrieve the pressed buttons
        unsigned int offset = 0;       // holds the display position
        unsigned int leftButton;       // holds the pressed button (default left gear)
        unsigned int rightButton;      // holds the pressed button (default right speed)
        unsigned long timeButton;      // holds the first pressed button time
        
        byte speeddata = 0;         // marker that new data are available
        byte geardata = 0;          // marker that new data are available   
        byte rpmdata = 0;           // marker that new data are available
        byte oildata = 0;           // marker that new data are available
        byte waterdata = 0;           // marker that new data are available
        byte fueldata = 0;           // marker that new data are available
        byte lapdata = 0;           // marker that new data are available
        byte positiondata = 0;           // marker that new data are available
        
        // char* neutral = "n";        // sets the character for neutral
        char* neutral = "0";        // sets the character for neutral
        char* reverse = "r";        // sets the character for reverse
           
        char* namesz = "88888888"; // sets a custom logo start up banner
       
        ////////////////////////////////////////////////////////////////////////////////
        // INITIALIZATION
        ////////////////////////////////////////////////////////////////////////////////
        void setup()
        {
          // serial initialization
          Serial.begin(115200);

          timeButton = millis();
         
          // initialize the screen with leds test
          module.clear();                       // clears the display from garbage if any
          module.setChars(namesz);           // prints the banner
          delay(1000);                                 // small delay 1 sec
          module.clear();                       // clears the display
          module.setLEDs(0b11111111 , 0b00000000); // all green leds
          delay(1000);                                 // small delay 1 sec
          module.setLEDs(0b00000000 , 0b11111111); // all reds leds
          delay(1000);                                 // small delay 1 sec
          module.setLEDs(0b00000000 , 0b00000000); // turn off all leds
         
          // read saved display config
          valButton = EEPROM.read(1020);
          rightButton = valButton;
          valButton = EEPROM.read(1021);
          leftButton = valButton;
          // if EEPROM empty
          if (rightButton == 255 && leftButton == 255){
            leftButton = btnGear; 
            rightButton = btnSpeed;
          }
          // show saved display config
          butDisp(leftButton,0);
          butDisp(rightButton,7);
          delay(1000);
          module.clear();           
        }
       
        ////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////// Main Loop ////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////
        void loop()
        ////////////////////////////////////////////////////////////////////////////////
        {

          // read input data
          readSerialData();
          
          // read pressed button
          buttons();

          // update dashboard
          dashboard();

        }
       
        ////////////////////////////////////////////////////////////////////////////////
        // Procedure: read serial data
        ////////////////////////////////////////////////////////////////////////////////
        void readSerialData()
        ////////////////////////////////////////////////////////////////////////////////
        {
          byte Data[5]={'0','0','0','0','0'};

          // parse the buffer
          if (Serial.available()>4){
            Data[0]=Serial.read();                   // read command
            if (Data[0]=='M'){                       // if rpm data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              rpm = NormalizeData(Data);
              rpmdata=1;                             // we got new data!       
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }
            if (Data[0] == 'S' ){                    // if speed data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              carspeed = NormalizeData(Data);
              speeddata=1;                           // we got new data!
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }
            if (Data[0] == 'G' ){                    // if gear data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              gear = NormalizeData(Data);
              geardata=1;                            // we got new data!
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }             
            if (Data[0] == 'O' ){                    // if oil data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              oil = NormalizeData(Data);
              oildata=1;                             // we got new data!
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }             
            if (Data[0] == 'W' ){                    // if water data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              water = NormalizeData(Data);
              waterdata=1;                           // we got new data!
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }             
            if (Data[0] == 'F' ){                    // if fuel data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              fuel = NormalizeData(Data);
              fueldata=1;                            // we got new data!
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }             
            if (Data[0] == 'L' ){                    // if lap data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              lap = NormalizeData(Data);
              lapdata=1;                             // we got new data!
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }             
            if (Data[0] == 'P' ){                    // if position data
              Data[1]=Serial.read();                 // store high byte
              Data[2]=Serial.read();                 //
              Data[3]=Serial.read();                 //
              Data[4]=Serial.read();                 // store low byte
              posit = NormalizeData(Data);
              positiondata=1;                        // we got new data!
              if (Serial.available()>4){
                Data[0]=Serial.read();               // re-read command
              }
            }             
            if (Data[0] == 'Z' ){                      // if Shutdown - Output
              Serial.flush();
              rpmmax=5000;                             // reset rpmmax
              module.clear();                          // clear display   
              module.setLEDs(0b00000000 , 0b00000000); // clear leds
            }             
          }
          if (Serial.available()>60) Serial.flush();
        }

        ////////////////////////////////////////////////////////
        // Procedure: display values
        ////////////////////////////////////////////////////////
        void dashboard()
        ////////////////////////////////////////////////////////
        {   
          if (speeddata == 1) {
            // filter of negative value
            // reported from 65535 to down in some games
            if (carspeed > 50000){                            // reverse velocity in some games are negative value
              carspeed = 65536 - carspeed;                    // convert positive with 16bit value
            }
            if (leftButton == btnSpeed) {
              module.setNumberPad(carspeed,2,3);              // displays numerical the speed
            }
            if (rightButton == btnSpeed) {
              module.setNumberPad(carspeed,7,4);              // displays numerical the speed
            }            
            // if (carspeed == 0){
            //   module.setChars("0",7);
            // }            
            speeddata=0;           
          } 
          if (geardata == 1) {
            if (leftButton == btnGear) {
              offset = 0;
            }
            if (rightButton == btnGear) {
              offset = 7;
            }            
            if (leftButton == btnGear || rightButton == btnGear) {
              if (gear >= 1 and gear <10 ){         
                module.setNumber(gear,offset);                        // displays numerical value of the current gear
              }
              if (gear == 0){
                module.setChars(neutral,offset);                      // displays the character for neutral
              }
              if (gear == (65535)){                                   // -1 that reprecents reverse rollover so...
                module.setChars(reverse,offset);                      // displays the character for reverse
              }
            }
            geardata=0;
          } 
          if (rpmdata == 1) {
            if (leftButton == btnRPM) {
              module.setNumberPad(rpm,3,4);
            }
            if (rightButton == btnRPM) {
              module.setNumberPad(rpm,7,5);
            }            
            // auto set max rpm
            // with filter of negative value
            // reported from 65535 to down in some games
            if (rpm>rpmmax && rpm<50000) {
              rpmmax=rpm;
            }
            rpmleds = map(rpm,0,rpmmax*1.1,0,9);              // distributes the rpm level to the 8 leds + 1 for shift change
            if (rpmleds==0){
              module.setLEDs(0b00000000 , 0b00000000);
            }
            if (rpmleds==1){
              module.setLEDs(0b00000001 , 0b00000000 );
            }
            if (rpmleds==2){
              module.setLEDs(0b00000011 , 0b00000000 );
            }
            if (rpmleds==3){
              module.setLEDs(0b00000111 , 0b00000000 );
            }
            if (rpmleds==4){
              module.setLEDs(0b00001111 , 0b00000000);
            }
            if (rpmleds==5){
              module.setLEDs(0b00011111 , 0b00000000);
            }
            if (rpmleds==6){
              module.setLEDs(0b00011111 , 0b00100000 );
            }
            if (rpmleds==7){
              module.setLEDs(0b00011111 , 0b01100000 );
            }
            if (rpmleds==8){
              module.setLEDs(0b11111111 , 0b1111111111 );
            }
            rpmdata=0;
          }
          if (oildata == 1) {
            if (leftButton == btnOIL) {
              module.setNumberPad(oil,2,3);
            }
            if (rightButton == btnOIL) {
              module.setNumberPad(oil,7,3);
            }            
            oildata=0;           
          } 
          if (waterdata == 1) {
            if (leftButton == btnWater) {
              module.setNumberPad(water,2,3);
            }
            if (rightButton == btnWater) {
              module.setNumberPad(water,7,3);
            }            
            waterdata=0;           
          } 
          if (fueldata == 1) {
            if (leftButton == btnFuel) {
              module.setNumberPad(fuel,1,2);
            }
            if (rightButton == btnFuel) {
              module.setNumberPad(fuel,7,2);
            }            
            fueldata=0;           
          } 
          if (lapdata == 1) {
            if (leftButton == btnLAP) {
              module.setNumberPad(lap,1,2);
            }
            if (rightButton == btnLAP) {
              module.setNumberPad(lap,7,2);
            }            
            lapdata=0;
          } 
          if (positiondata == 1) {
            if (leftButton == btnPosition) {
              module.setNumberPad(posit,1,2);
            }
            if (rightButton == btnPosition) {
              module.setNumberPad(posit,7,2);
            }            
            positiondata=0;
          } 
        }

        ////////////////////////////////////////////////////////
        // Procedure: read pushed button
        ////////////////////////////////////////////////////////
        void buttons()
        ////////////////////////////////////////////////////////
        {        
          valButton = module.getButtons();
          if (valButton > 0 && valButton != beforeButton){
            // if second button
            if (millis()-timeButton < 3000){
              // store slection
              leftButton = valButton;
              EEPROM.write(1021,leftButton);
              //
              butDisp(leftButton,0);
              delay(300);
              module.clear();
            }
            // if first button
            if (millis()-timeButton > 3000){
              module.clear(); 
              // store selection
              rightButton = valButton;
              EEPROM.write(1020,rightButton);              
              leftButton = 0;
              EEPROM.write(1021,leftButton);              
              //
              butDisp(rightButton,7);
              delay(300);
              module.clear(); 
              timeButton = millis();              
            }
          }
          beforeButton = valButton;
        }
        
        ////////////////////////////////////////////////////////////////////////////////
        // Function: display select function
        ////////////////////////////////////////////////////////////////////////////////
        void butDisp(int iButton,int off){
        ////////////////////////////////////////////////////////////////////////////////
              if (iButton == btnSpeed){
                  module.setChars("S",off);
              }
              if (iButton == btnGear){
                  module.setChars("G",off);
              }
              if (iButton == btnRPM){
                  module.setChars("M",off);
              }
              if (iButton == btnOIL){
                  module.setChars("O",off);
              }
              if (iButton == btnWater){
                  module.setChars("W",off);
              }
              if (iButton == btnFuel){
                  module.setChars("F",off);
              }
              if (iButton == btnLAP){
                  module.setChars("L",off);
              }
              if (iButton == btnPosition){
                  module.setChars("P",off);
              }
        }
        
        ////////////////////////////////////////////////////////////////////////////////
        // Function: convert Hex to Dec
        ////////////////////////////////////////////////////////////////////////////////
        unsigned int NormalizeData(byte x[5])
        ////////////////////////////////////////////////////////////////////////////////
        {
        unsigned int result;
  
        for (int i=1; i<5; i++)
        {
          if (x[i]>47 && x[i]<58 ){   //for xA to xF
            x[i]=x[i]-48;
          }                       
          if (  x[i]>64 && x[i]<71 ){ //for x0 to x9
            x[i]=x[i]-55;               
          }
        }    
        result = (x[1]*16*16*16+x[2]*16*16+x[3]*16+x[4]);
        return result;
        }
        
